from taggingbackends.data.labels import Labels
from taggingbackends.features.skeleton import get_5point_spines
from maggotuba.models.trainers import MaggotTrainer, MultiscaleMaggotTrainer, new_generator
import numpy as np

def predict_model(backend, **kwargs):
    """
    This function generates predicted labels for all the input data.

    The input files can be read from any directory.
    All generated/modified files should be written to `data/interim` or
    `data/processed`.
    The predicted labels are expected in `data/processed`.

    The `predict_model.py` script is required.
    """
    # we pick files in `data/interim` if any, otherwise in `data/raw`
    input_files = backend.list_interim_files()
    if not input_files:
        input_files = backend.list_input_files()
    assert 0 < len(input_files)
    # initialize output labels
    input_files, labels = backend.prepare_labels(input_files)
    assert 0 < len(input_files)
    # load the model
    model_files = backend.list_model_files()
    config_file = [file for file in model_files if file.name.endswith("config.json")]
    n_config_files = len(config_file)
    if n_config_files == 0:
        raise RuntimeError(f"no such tagger found: {backend.model_instance}")
    config_file = [file
            for file in config_file
            if file.name.endswith("clf_config.json")
            and file.parent == backend.model_dir()]
    assert len(config_file) == 1
    config_file = config_file[-1]
    if 2 < n_config_files:
        model = MultiscaleMaggotTrainer(config_file)
    else:
        model = MaggotTrainer(config_file)
    #
    labels.labelspec = model.clf_config["behavior_labels"]
    #
    if len(input_files) == 1:
        file = input_files[0]
        if file.name.startswith("larva_dataset_") and file.name.endswith(".hdf5"):
            ret = predict_larva_dataset(backend, model, file, labels, **kwargs)
            return labels if ret is None else ret
    #
    ret = predict_individual_data_files(backend, model, input_files, labels)
    return labels if ret is None else ret

def predict_individual_data_files(backend, model, input_files, labels):
    from taggingbackends.data.trxmat import TrxMat
    from taggingbackends.data.chore import load_spine
    import taggingbackends.data.fimtrack as fimtrack
    #
    _break = False # for now, a single file can be labelled at a time
    for file in input_files:
        # load the input data (or features)
        if _break:
            print(f"ignoring file: {file.name}")
            continue
        elif file.name.endswith(".spine"):
            spine = load_spine(file)
            run = spine["date_time"].iloc[0]
            larvae = spine["larva_id"].values
            t = spine["time"].values
            data = spine.iloc[:,3:].values
        elif file.name.endswith(".mat"):
            trx = TrxMat(file)
            t = trx["t"]
            data = trx["spine"]
            run, data = next(iter(data.items()))
            if run == "spine":
                run, data = next(iter(data.items()))
            t = t[run]
        elif file.name.endswith(".csv"):
            if labels.camera_framerate:
                print(f"camera frame rate: {labels.camera_framerate}fps")
            else:
                print("assuming 30-fps camera frame rate")
                labels.camera_framerate = 30
            t, data = fimtrack.read_spines(file, fps=labels.camera_framerate)
            run = "NA"
        else:
            print(f"ignoring file: {file.name}")
            continue
        # downsample the skeleton
        if isinstance(data, dict):
            for larva in data:
                data[larva] = get_5point_spines(data[larva])
        else:
            data = get_5point_spines(data)
        # assign labels
        if isinstance(data, dict):
            ref_length = np.median(np.concatenate([
                model.body_length(spines) for spines in data.values()
                ]))
            model.average_body_length = ref_length
            print(f"average body length: {ref_length}")
            for larva, spines in data.items():
                predictions = model.predict((t[larva], spines))
                if predictions is None:
                    print(f"failure at windowing track: {larva}")
                else:
                    labels[run, larva] = dict(zip(t[larva], predictions))
        else:
            ref_length = np.median(model.body_length(data))
            model.average_body_length = ref_length
            print(f"average body length: {ref_length}")
            for larva in np.unique(larvae):
                mask = larvae == larva
                predictions = model.predict((t[mask], data[mask]))
                if predictions is None:
                    print(f"failure at windowing track: {larva}")
                else:
                    labels[run, larva] = dict(zip(t[mask], predictions))
        # save the predicted labels to file
        labels.dump(backend.processed_data_dir() / "predicted.label")
        #
        _break = True

def predict_larva_dataset(backend, model, file, labels, subset="validation"):
    from taggingbackends.data.dataset import LarvaDataset
    #
    dataset = LarvaDataset(file, new_generator())
    return model.predict(dataset, subset)


from taggingbackends.main import main

if __name__ == "__main__":
    main(predict_model)

