import numpy as np
import torch
import torch.nn as nn
from behavior_model.models.neural_nets import device
#import behavior_model.data.utils as data_utils
from maggotuba.models.modules import SupervisedMaggot, MultiscaleSupervisedMaggot
from taggingbackends.features.skeleton import interpolate

"""
This model borrows the pre-trained MaggotUBA encoder, substitute a dense layer
for the decoder, and (re-)trains the entire model.

Attribute `config` refers to MaggotUBA autoencoder.
Attribute `clf_config` refers to the supervised model, both the retrained
encoder and the trained classifier.

Several data preprocessing steps are included for use in prediction mode.
Training the model instead relies on the readily-preprocessed data of a
*larva_dataset hdf5* file.
"""
class MaggotTrainer:
    def __init__(self, cfgfilepath, behaviors=[], n_layers=1,
            average_body_length=1.0, device=device):
        self.model = SupervisedMaggot(cfgfilepath, behaviors, n_layers)
        self.average_body_length = average_body_length # usually set later
        self.device = device

    @property
    def config(self):
        return self.model.encoder.config

    @property
    def clf_config(self):
        return self.model.clf.config

    @property
    def labels(self):
        return self.model.clf.behavior_labels

    @labels.setter
    def labels(self, labels):
        self.model.clf.behavior_labels = labels

    @property
    def swap_head_tail(self):
        return self.config.get('swap_head_tail', True)

    @swap_head_tail.setter
    def swap_head_tail(self, b):
        self.config['swap_head_tail'] = b

    ### TODO: move parts of the below code in a features module
    # all the code in this section is called by `predict` only
    def window(self, t, data):
        interpolation_args = {k: self.config[k]
                              for k in ('spine_interpolation', 'frame_interval')
                              if k in self.config}
        winlen = self.config["len_traj"]
        N = data.shape[0]+1
        if interpolation_args:
            for m in range(0, N-1):
                win = interpolate(t, data, m, winlen, **interpolation_args)
                if win is not None:
                    assert win.shape[0] == winlen
                    yield win
        else:
            for m in range(0, N-winlen):
                n = m + winlen
                yield data[m:n]

    def pad(self, data):
        winlen = self.config["len_traj"]
        if data.shape[0] == 1:
            return data
        else:
            ind = np.r_[
                    np.zeros(winlen // 2, dtype=int),
                    np.arange(data.shape[0]),
                    (data.shape[1]-1) * np.ones(winlen // 2 - 1, dtype=int),
                    ]
            return data[ind]

    def body_length(self, data):
        dx = np.diff(data[:,0::2], axis=1)
        dy = np.diff(data[:,1::2], axis=1)
        return np.sum(np.sqrt(dx*dx + dy*dy), axis=1)

    def normalize(self, w):
        # center coordinates
        wc = np.mean(w[:,4:6], axis=0, keepdims=True)
        w = w - np.tile(wc, (1, 5))
        # rotate
        v = np.mean(w[:,8:10] - w[:,0:2], axis=0)
        v = v / np.sqrt(np.dot(v, v))
        c, s = v / self.average_body_length # scale using the rotation matrix
        rot = np.array([[ c, s],
                        [-s, c]]) # clockwise rotation
        w = np.einsum("ij,jkl", rot, np.reshape(w.T, (2, 5, -1), order='F'))
        return w

    def preprocess(self, t, data):
        ws = []
        for w in self.window(t, data):
            ws.append(self.normalize(w))
        if ws:
            ret = self.pad(np.stack(ws))
            if self.swap_head_tail:
                ret = ret[:,:,::-1,:]
            return ret
    ###

    def forward(self, x, train=False):
        if train:
            return self.model(x)
        else:
            if isinstance(x, torch.Tensor):
                if x.dtype is not torch.float32:
                    x = x.to(torch.float32)
            else:
                x = torch.from_numpy(x.astype(np.float32))
            y = self.model(x.to(self.device))
            return y.cpu().numpy()

    def prepare_dataset(self, dataset):
        try:
            dataset.batch_size
        except AttributeError:
            dataset.batch_size = self.config["batch_size"]
        target_winlen = self.config["len_traj"]
        # 0-indexed midpoint -> first point of the second half if length is even,
        #                       middle point if length is odd
        midpoint = dataset.window_length // 2
        # to help reason about indices, we want:
        # * for a 2-point window: the midpoint and the point before,
        # * for a 3-point window: the midpoint, the point before and the point after,
        # * and so on.
        before = target_winlen // 2
        after = target_winlen // 2 + target_winlen % 2
        assert before + after == target_winlen
        if not (0 <= midpoint - before and midpoint + after <= dataset.window_length):
            raise ValueError(f"the dataset can provide segments of up to {dataset.window_length} time points")
        dataset._mask = slice(midpoint - before, midpoint + after)

    def train(self, dataset):
        self.prepare_dataset(dataset)
        model = self.model
        model.train() # this only sets the model in training mode (enables gradients)
        model.to(self.device)
        criterion = nn.CrossEntropyLoss()
        # pre-train the classifier with static encoder weights
        optimizer = torch.optim.Adam(model.clf.parameters())
        print("pre-training the classifier...")
        for step in range(self.config["optim_iter"] // 2):
            optimizer.zero_grad()
            # TODO: add an option for renormalizing the input
            data, expected = self.draw(dataset)
            predicted = self.forward(data, train=True)
            loss = criterion(predicted, expected)
            #print("pre-train", torch.mean(loss).detach().numpy())
            loss.backward()
            optimizer.step()
        # fine-tune both the encoder and the classifier
        optimizer = torch.optim.Adam(model.parameters())
        print("fine-tuning the encoder and classifier...")
        for step in range(self.config["optim_iter"] // 2):
            optimizer.zero_grad()
            data, expected = self.draw(dataset)
            predicted = self.forward(data, train=True)
            loss = criterion(predicted, expected)
            #print("fine-tune", torch.mean(loss).detach().numpy())
            loss.backward()
            optimizer.step()
        #
        return self

    def draw(self, dataset, subset="train"):
        data, expected = dataset.getobs(subset)
        if isinstance(data, list):
            data = torch.stack(data)
        data = data.to(torch.float32).to(self.device)
        if isinstance(expected, list):
            expected = torch.stack(expected)
        if subset.startswith("train"):
            expected = expected.to(torch.long).to(self.device)
        return data, expected

    @torch.no_grad()
    def predict(self, data, subset=None):
        model = self.model
        model.eval()
        model.to(self.device)
        if subset is None:
            # data is a (times, spines) couple
            data = self.preprocess(*data)
            if data is None:
                return
            output = self.forward(data)
            label_ids = np.argmax(output, axis=1)
            labels = [self.labels[label] for label in label_ids]
            return labels
        else:
            dataset = data
            self.prepare_dataset(dataset)
            predicted, expected = [], []
            for data, exp in dataset.getsample(subset, "all"):
                output = self.forward(data)
                pred = np.argmax(output, axis=1)
                exp = exp.numpy()
                assert pred.size == exp.size
                predicted.append(pred)
                expected.append(exp)
            return np.concatenate(predicted), np.concatenate(expected)

    def save(self):
        self.model.save()

def new_generator(seed=0b11010111001001101001110):
    return torch.Generator(device).manual_seed(seed)

class MultiscaleMaggotTrainer(MaggotTrainer):
    def __init__(self, cfgfilepath, behaviors=[], n_layers=1,
            average_body_length=1.0, device=device):
        self.model = MultiscaleSupervisedMaggot(cfgfilepath, behaviors, n_layers)
        self.average_body_length = average_body_length # usually set later
        self.device = device
        self._default_encoder_config = None
        # check consistency
        ref_config = self.config
        for attr in ["batch_size", "optim_iter"]:
            # TODO: add pretraining_iter and finetuning_iter parameters in
            #       clf_config to have a lever other than optim_iter, that
            #       could consequently be ignored
            for enc in self.model.encoders:
                assert enc.config[attr] == ref_config[attr]

    @property
    def config(self):
        if self._default_encoder_config is None:
            len_traj = 0
            for enc in self.model.encoders:
                cfg = enc.config
                if len_traj < cfg["len_traj"]:
                    len_traj = cfg["len_traj"]
                    self._default_encoder_config = cfg
        return self._default_encoder_config

